﻿/*	Copyright: autocart (user name at the ahk and XYplorer forums) unless otherwise noted in the code
Version: 4.0
Date: 2020-05-01

General description:

Preamble:
Most of the code was written years ago. Looking at the code now, it does not look pretty at all. I feel a need to rewrite it from scratch but don't have the time nor the motivation for it all at the moment. For now, I just started a little bit and mostly changed the things, which seemed to me most important. In a later version I might rewrite the code - or not.

Purpose and usage of the script:
This script works in combination with the file explorer XYplorer (from now on called "XY") and Windows Explorer (from now on called "WE").
It was written for the following scenarios:

(1)	XY's setting "XY is default file manager" not getting respected:
	Despite XY's build-in optional functionality to act as the system's default file manager (XY menu Tools/Configuration/Other/Shell Integration), sometimes there still are WE windows opening instead of XY (e.g. from within certain 3rd-party-software or for other reasons).

(2)	List items not getting automatically selected in XY:
	At other times XY *is* sucessfully used as the system's default file manager to open a folder (if that setting is on) but any list itmes which would normally be selected in WE do *not* get selected in XY (e.g. when clicking in Firefox in the "downloads" window on the "open containing folder" button).

For such scenarios, this script, if running in the system tray, catches any newly opening WE window, reads the path and selected items shown in that WE window, closes that WE window, opens a new tab in XY with that path *and* selects any list items in XY that where selected in WE before (and scrolling them into view), thus "redirecting" the WE window to XY.

The script will stay in the system tray until manually exited through the script's tray context menu.

Some of the scripts settings can be easily customized by the user. These settings are located in a separate config file, named "redirectWEwindowsToXY_v#.config.ahk", with "#" standing for the *main* version numer of the script which it is compatible to.

Hint:
For the script to work also in cases of scenario No. (2), of course, the XY setting to act as the system's default file manager must be turned *off*.

More details:

If one wants any WE windows to be kept as WE windows (and *not* be redirected to XY) - e.g. control panel stuff - then there are these options:

(1)	Control-key-escape:
	Hold the left or right control key pressed down while the WE window opens, which you want to keep. Typically, though, the control key must be held down until the WE window has actually opened. If released too early, it might not work as intended.

	Example:
	With the script running, pressing Win+E would open a new WE window which normally would get redirected to XY.
	Pressing Control+Win+E, the new WE window would *not* get redirected to XY. (Of course the Control-key-escape option should work also if a WE window is opened by other means.)

(2)	Pre-defined keywords / paths:
	The script allows for pre-defined keywords and paths, which are compared to the location being displayed in the WE window (typically the content of the WE address bar, e.g. "Computer", "Libraries", "Pictures" ... and paths without trailing backslash). These keywords and paths can be either set in the variable vWindowsExplorerPaths in the separate config file "redirectWEwindowsToXY_v#.config.ahk" - separated by the pipe character "|". Else they can also be passed to the script as command line parameters - separated by spaces (and quoted if they contain spaces). Both kinds of definition *are* allowed together at the same time.

	Example for the command line parameters option:
	In case of command line parameters (or parameters passed in the target line of a windows shortcut file) the command line could look like this:
	"path_to_autohotkey.exe" "path_to_script.ahk" Computer Libraries Pictures "C:\some kind of folder\another folder"

	Hint:
	In the separate config file "redirectWEwindowsToXY_v#.config.ahk" there are some suggestions for keywords to be used on an English version of Windows.

	The variable vKeepOrRedirect_matchingWEPaths in the separate config file "redirectWEwindowsToXY_v#.config.ahk" determines what will happen in case of a match (when comparing the WE window's location with the keywords/paths definitions). It must be set to either "Keep" or "Redirect". If vKeepOrRedirect_matchingWEPaths is set to "Keep", then the matching WE windows are kept as WE windows (and the others are redirected to XY). If vKeepOrRedirect_matchingWEPaths is set to "Redirect" then the matching WE windows are redirected to XY (and the others are kept as WE windows).
	
	The method for comparing the WE window's location with the keywords/paths definitions, can either be set to use Regular Expressions (RegEx), which is recommended due to its ability to formulate general patterns. To do so, the variable vUseRegExForPathCompare in the separate config file "redirectWEwindowsToXY_v#.config.ahk" must be set to 1. Else, if vUseRegExForPathCompare is set to 0, then - to be a match - a keyword or path and the WE window's location must be identical.
	
	Hint:
    Of course, if RegEx is used as a compare method, then the keywords and paths must be specified using RegEx syntax.
	Here are some helpful sites for learning RegEx (PCRE flavor):
	https://www.regular-expressions.info/quickstart.html
	https://regex101.com/
	
	If RegEx is used as a compare method, then any additional keywords/paths specified in the extra variable vExceptionPathsInCaseOfRegEx in the separate config file "redirectWEwindowsToXY_v#.config.ahk" act as a subset of exceptions to the main keywords/paths definitions specified in vWindowsExplorerPaths. Therefore, the main definitions in vWindowsExplorerPaths can contain a more general pattern and the extra variable vExceptionPathsInCaseOfRegEx can contain exceptions from that general pattern. (This is *only* available for the compare method RegEx.)

Even more details:

WE windows, which show the content of a zip file, are automatically kept as WE windows if the "identical" compare method is used. This is, because XY does not support this kind of view and the "identical" compare method is not capable of defining zip files as a pattern. However, if RegEx is used as compare method, then the user must specify it manually in the RegEx pattern if he/she wants a zip file-content-view to be kept in WE (e.g. with \.zip$ either as part of a Keep-pattern or as part of a exeptionPaths-when-Redirect-pattern).

If XY is not running it will be started automatically if the path to XYplorer.exe is specified correctly. The path can be specified in the variable vPathToXYplorerForAutostart in the separate config file "redirectWEwindowsToXY_v#.config.ahk". A tooltip close to the mouse cursor will indicate the starting process of XY. If the XY-autostart should not work for some reason (e.g. if the wrong or no path is specified) the script will display a message box and wait for the user to start XY manually.

If the user wishes for the script to apply auto-redirection also to all WE windows already existing at the moment of the script's start (not just to newly created ones while it is running) then this can be activated in the variable vApplyRulesToExistingWEwindowsAtScriptStart in the separate config file "redirectWEwindowsToXY_v#.config.ahk". Setting the variable to 1 will activate that behaviour. Setting it to 0 will deactivate it.

In order to manually redirect a given WE window to XY by force (regardless of the rules for automatic redirection), the user can specify a hotkey in the variable vHotkeyInWEwindowsToTriggerRedirectionToXY in the separate config file "redirectWEwindowsToXY_v#.config.ahk". (A hotkey refers to pressing a combination of keys on the keyboard, namely a character key together with one or more modifier keys. Modifier keys are Control, Alt, Shift and the Windows key.) The WE window will be redirected if the hotkey is pressed while the WE window is active. More details on how to specify the hotkey are given right next to the variable itself. This hotkey can be activated or deactivated on its own (just in case) in the script's system tray context menu. If no such hotkey is specified, then the system tray context menu item will not exist.

The whole script with all its functionality can be suspended in the script's system tray context menu or with a global hotkey, which can be specified in the variable vHotkeyGlobalToSuspendUnsuspendTheWholeScript in the separate config file "redirectWEwindowsToXY_v#.config.ahk".

Known limitations:

-)	Whenever a WE window is "redirected" to XY, it will still open for a short moment in a normal way and immediately be closed again. For this script to work, I do not know of any way how to avoid that.
-)	In case of multiple XY windows being open at the same time, the most recently active XY window will be used for redirection.
-)	This script has not been tested with WE add-ons, which add tabs to WE.

For suggestions and ideas for improvement I am grateful.

Changes in version 3.0:

-)	Improved reliability:
	Before, the script sometimes missed WE windows or seemingly could not retrieve its path and then did not react as expected.
-)	Improved behaviour when paused:
	Now the script is respecting the script's paused state (see the script's system tray context menu) in that it stops redirecting when the script paused.
	Before it ignored the state of being paused and redirected anyway.
-)	Changed the behaviour of the script regarding already existing WE windows:
	In previous versions, when the title bar of already existing WE windows (which existed prior to the scripts start) changed (e.g. when changing a folder in such a window or when refreshing by hitting F5), then the script also used to redirect such WE windows to XY.
	Not anymore. Now, if a WE window already exist at the scripts start, the script leaves it untouched.
-)	Added a Control-key-escape option:
	If the left or right Control key is pressed while a WE window opens, then this particular WE window will *not* get redirected to XY.

Changes in version 3.1:

-)	Fixed the problem that the script would crash if XY was not active.
-)	The path to XY can now be specified in the variable $pathToXYplorer at the beginning of the code.
-)	In case of a XY autostart process, a tooltip will be diplayed close to the mouse cursor.
-)	In case the XY autostart process could not be started, a message box will pop.

Changes in version 3.2:

-)	Added code for an optional custom tray icon and supplied a simple icon with it.
-)	Renamed the variables; especially the variable with all the keywords and paths is now called vWindowsExplorerPaths.
-)	Added the options/variables vKeepOrRedirect_matchingWEPaths, vUseRegExForPathCompare and vExceptionPathsInCaseOfRegEx.
-)	Added some visual marking of the user customize section in the code below
-)	In case the variable vKeepOrRedirect_matchingWEPaths has an incorrect value, a message box will pop.
-)	Changed the hard coded behaviour of *not* redirecting the display of the content of zip files to XY: Now it is only hard coded for the case if RegEx is *not* used as compare method. If RegEx is used, then the behaviour for zip files can be specified in the RegEx pattern.

Changes in version 3.3:

-)	Fixed some wrong logic in the processing of events.
	Previously, if XY was not already running, it would be automatically started also if the WE window was to be kept as WE window. In that scenario also, if no correct path to XYplorer.exe was specified, then the script would still pop the corresponding message box and wait for XY to be started manually. This made not even any sense, since the WE window would anyway not be redirected to XY. This is changed now. The script will only check for the open XY window and react accordingly if the WE window is to be redirected.
-)	Fixed a logical bug with RegEx compare.
	If hard coded keywords/paths were used AND command line parameters AND RegEx was used as compare method, then there was a logical bug which could have led to everything being a match (besides the RegEx-exeptions)
-)	Changed the order of the help text a bit.
	Adjusted the order of the variables in the "USER CUSTOMIZE SECTION" of the code accordingly.
-)	Added the variable vApplyRulesToExistingWEwindowsAtScriptStart to optionally apply the specified rules also to existing WE windows at script start.
-)	Added the variable vHotkeyInWEwindowsToTriggerRedirectionToXY to be able to specify a hotkey to redirect WE windows to XY manually, if pressed while the targeted WE window is active.
-)	Simplified the system tray context menu to three items:
	(1) Toggle menu item to use/activate the specified hotkey for redirecting WE windows manually (or deactivate it).
	(2) Toggle menu item to Pause / Suspend Script which both pauses the script and suspends the hotkeys.
	(3) Fully exit the script.

Changes in version 3.4:

-)	Maybe fixed that XY sometimes crashed when being autostarted.
	When XY was autostarted it sometimges crashed, as it seems, without the line "WinWait, XYplorer ahk_class ThunderRT6FormDC", which is strange to me because the code before that line should have aleady made sure that the XY window exists. However, it did not always crash. Re-added that line in for now, nevertheless. So far, it never crashed on autostart with that line.
-)	Fixed / Changed (hopefully improved) the way how paths, which are being redirected to XY, are passed to XY.
	There could have been problems with how paths were compared to the existing tabs. Especially if the redirected path was a sub path of an existing tab, it would have not opened a new tab but changed the current tab to the new path.
-)	Fixed a bug by which the script would send XY into a scripting error, if the redirected path contained single quote characters.
-)	Fixed a bug that if a path was redirected to XY and the list in XY did *not* have focus, then any selected items in WE would be selected in XY but not be scrolled into view. Now they are (respectively the first selected item is).
-)	Fixed a bug preventing to manually force redirect zip content views to XY.
-)	Fixed / Changed the way how paths, that are really files (e.g zip files), are handled when passed to XY.
	There was an inconsistency in the code how such paths were handled, which could have led to XY displaying nothing in a new tab. This was highly unlikely to surface to the user, but might have after the fix above. Now, if the path is really a file, then the path to the parent directory of that file is always taken as the redirected path and the file as selection.

Changes in version 3.4.1:

-)	Fixed a bug with none-ascii characters. Saved the script file as with UTF-8 *with* BOM.
	Before it was saved as UTF-8 without BOM, which caused none-ascii characters to be processed incorrectly.

Changes in version 4.0:

-)	Moved the user customize section out into a separate .config.ahk file.
	This helps the user to not have to update the user customizeation part of the script each time a new version is released. Plus, it is easier to keep the overview in a much shorter file without the distraction of the intro or the main code below.
	The new .config.ahk file carries the *main* version number of the script version, which it belongs to, in its name. The *main* version number of the script (and thus the name of the .config.ahk file) will from now on be incremented only, if the compatibility of the .config.ahk file changes. Thus it is easy to see, which script needs which version of the .config.ahk file - just look at the script's *main* version number.
-)	Added a check for the script file's encoding and the encoding of the new config.ahk file.
	If they are not saved in the UTF-8 format with BOM, then the script will pop a message box and refuse to run.
-)	Added a global hotkey, specified in the variable vHotkeyGlobalToSuspendUnsuspendTheWholeScript in the .config.ahk file, which will suspend or unsuspend the script.
-)	Added a tray menu item which toggles the use of Ctrl to manually keep a WE window (just in case).
-)	Renamed the tray menu items a bit.

Possible ideas for the future:

-)	make a separate help/readme file
-)	maybe handle also other message scenarios - not just "window created"
-)	maybe re-introduce checking of paths in existing WE windows while browsing the file system in that WE window. I think, this has pros and cons. The pros are that if I specifially browse to a path which I want to open in XY it will be done automatically. The cons are that if I browse to a path which gets redirected to XY, I might have forgotten to think of it and might not want that at this very moment. I think for now, the new hotkey feature with which it is possible to manually force-redirect the current path/view in an WE window to XY, might be the best solution. But I am open for suggestions.
-)	hotkey(s) for pausing and un-pausing the script - either global or only when XY is active? - with the toggle-option in the tray menu
-)	improve the MsgBox popping when the WE path could not be determined (maybe with debug-information that a user could then post in the forum for my knowledge) and make it easier for the end user to disable that MsgBox if he/she would want to not see it
-)	implement an option somehow to "move" XY paths in the other direction to Windows Explorer. If at all, then this would probably only make sense with a manual trigger, not automatically for pre-defined paths, or???
*/



;;;******************************************************************************************
;;;*****************************   HERE STARTS THE CODE   ***********************************
;;;******************************************************************************************

#NoEnv  ;; Recommended for performance and compatibility with future AutoHotkey releases.
#SingleInstance force
#Persistent
; #Warn  ;; Enable warnings to assist with detecting common errors.
SendMode Input  ;; Recommended for new scripts due to its superior speed and reliability.
SetWorkingDir %A_ScriptDir%  ;; Ensures a consistent starting directory.


#Include, %A_ScriptDir%\redirectWEwindowsToXY_v4.config.ahk

;; Checks the specified files below for the correct file encoding "UTF-8" with BOM, if not compiled.
if !A_IsCompiled
{
	vFullPathFilesToBeCheckedForEncoding := A_ScriptFullPath . "|" . A_ScriptDir . "\redirectWEwindowsToXY_v4.config.ahk"
	Loop, Parse, vFullPathFilesToBeCheckedForEncoding, |
	{
		voFile := FileOpen(A_LoopField, "r")
		if (not voFile)
		{
			MsgBox, 262144, %A_ScriptName%, The attempt to check the file "%A_LoopField%" for the right encoding failed, because it could not be opened for reading, which is weird, because the AutoHokey exe itself had no problem reading it.`n`nPlease debug and then start the script again.`n`nThe script will terminate as soon as this dialog is closed.
			ExitApp
		}
		else if (not (voFile.Encoding == "UTF-8" and voFile.Position > 0))
		{
			if voFile.Position = 0
				vtextCurrentFileFormat := "However, it is saved without BOM. Die actual file format could thus not be determined."
			else
				vtextCurrentFileFormat := "However, it is saved as """ . voFile.Encoding . """."
			MsgBox, 262144, %A_ScriptName%, The file "%A_LoopField%" is saved in the wrong file format. It must be "UTF-8" *with* BOM. %vtextCurrentFileFormat%`n`nPlease save the file "%A_LoopField%" in "UTF-8" *with* BOM and then start the script again.`n`nThe script will terminate as soon as this dialog is closed.
			voFile.Close()
			ExitApp
		}
		voFile.Close()
	}
}


;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~   INITIALIZATION of TRAY MENU - START   ~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
if !A_IsCompiled
{
	Menu, Tray, UseErrorLevel
	Menu, Tray, Icon, %A_ScriptDir%\redirectWEwindowsToXY.ico
}

Menu, Tray, NoStandard
Menu, Tray, Add, Use [Ctrl] to manually keep a WE window, _ToggleUseControlToKeepWE ;_ToggleUseControlToKeepWE()
Menu, Tray, Check, Use [Ctrl] to manually keep a WE window
if (vHotkeyInWEwindowsToTriggerRedirectionToXY)
{
	vDisplayFormatOfHKIWETTRTXY := StrReplace(StrReplace(StrReplace(StrReplace(vHotkeyInWEwindowsToTriggerRedirectionToXY, "+", "Shift+"), "#", "Win+"), "!", "Alt+"), "^", "Ctrl+")
	Menu, Tray, Add, Use [%vDisplayFormatOfHKIWETTRTXY%] to manually redirect WE to XY, _ToggleHotkeyIWETTRTXY ;_ToggleHotkeyIWETTRTXY()
	Menu, Tray, Check, Use [%vDisplayFormatOfHKIWETTRTXY%] to manually redirect WE to XY
}
vDisplayFormatOfHKGTSUSTWS := StrReplace(StrReplace(StrReplace(StrReplace(vHotkeyGlobalToSuspendUnsuspendTheWholeScript, "+", "Shift+"), "#", "Win+"), "!", "Alt+"), "^", "Ctrl+")
Menu, Tray, Add, Suspend Script ([%vDisplayFormatOfHKGTSUSTWS%]), _TogglePauseAndSuspendScript ;_TogglePauseAndSuspendScript()
Menu, Tray, Add, Exit Script, _ExitApp ;_ExitApp()
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~   INITIALIZATION of TRAY MENU - END   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


if (vHotkeyInWEwindowsToTriggerRedirectionToXY)
{
	Hotkey, IfWinActive, ahk_class CabinetWClass
	Hotkey, %vHotkeyInWEwindowsToTriggerRedirectionToXY%, _TriggerWEwindowRedirectionToXY ;_TriggerWEwindowRedirectionToXY()
	Hotkey, If
}

if (vHotkeyGlobalToSuspendUnsuspendTheWholeScript)
{
	Hotkey, %vHotkeyGlobalToSuspendUnsuspendTheWholeScript%, _TogglePauseAndSuspendScript ;_TogglePauseAndSuspendScript()
}


;; checks if vKeepOrRedirect_matchingWEPaths has an allowed value and, if not, pops a message box
if (not (vKeepOrRedirect_matchingWEPaths = "Keep" or vKeepOrRedirect_matchingWEPaths = "Redirect"))
{
	MsgBox, 262144, %A_ScriptName%, The variable vKeepOrRedirect_matchingWEPaths must be either "Keep" or "Redirect", however it is "%vKeepOrRedirect_matchingWEPaths%".`n`nPlease, correct it in the user customize section of the code.`n`nThe script will terminate as soon as this dialog is closed.
	ExitApp
}

vKeepIfCtrlHeldDown := 1

;; Before adding the command line parameters, ensure that the first and last char of vWindowsExplorerPaths is a pipe "|".
;; However, the end must be checked first if it is already a pipe, because if two consecutive pipes end up in the middle of the string (after command line parameters are added) then this would lead to unwanted results in case of the RegEx compare method. (compare below)
if (not SubStr(vWindowsExplorerPaths, 0) = "|")
	vWindowsExplorerPaths := "|" . vWindowsExplorerPaths . "|"
else
	vWindowsExplorerPaths := "|" . vWindowsExplorerPaths

/*	About incorrectly passed command line parameters:
	The help file says:
	Any parameter that contains spaces should be enclosed in quotation marks. A literal quotation mark may be passed in by preceding it with a backslash (\"). Consequently, any trailing slash in a quoted parameter (such as "C:\My Documents\") is treated as a literal quotation mark (that is, the script would receive the string C:\My Documents"). To remove such quotes, use StringReplace, 1, 1, ",, All.
	-> In my opinion that is a two edged sword. The core problem is that literal quotation marks are escaped with a backslash. Thus, two consecutive parameters of which the first ends on \" (backslash quote) are treated as one, no matter what. So, I comment the "correction" out in the next line for now.
	StringReplace, vWindowsExplorerPaths, vWindowsExplorerPaths, ",, All
	If wanted/needed, the whole raw command line can be correctly read using full_command_line := DllCall("GetCommandLine", "Str")
	Then it could be custom-parsed.
	For now, I go with instructing the user to NOT pass paths with a trailing backslash.
*/

For n, param in A_Args
{
    vWindowsExplorerPaths .= param . "|"
}

;; if RegEx is used as comparision method, then the pipes "|" at the beginning and end must be removed again, because otherwise the ends would be interpreted as empty patterns which would always match
if vUseRegExForPathCompare
	vWindowsExplorerPaths := Trim(vWindowsExplorerPaths, "|")


;; Make the GUI window the last found window for use by the line below.
Gui +LastFound
vHWnd := WinExist()
DllCall( "RegisterShellHookWindow", UInt, vHWnd )
vMsgNum := DllCall( "RegisterWindowMessage", Str, "SHELLHOOK" )
OnMessage( vMsgNum, "ShellMessage" )


if (vApplyRulesToExistingWEwindowsAtScriptStart)
{
	WinGet, vExistingCabinetWClassWindows, List, ahk_class CabinetWClass
	Loop %vExistingCabinetWClassWindows%
	{
		ShellMessage(1, vExistingCabinetWClassWindows%A_Index%)
	}
}

return

;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~   END of AUTOEXECUTE SECTION   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



^#e::
;; Using this construct so that the user in this particular case does not *have* to keep pressing the Ctrl key down until the WE window appears.
OnMessage( vMsgNum, "ShellMessage", 0 )
OnMessage( vMsgNum, "ShellMessage_forCtrlWinE" )
Send, #e
return


ShellMessage_forCtrlWinE( wParam, lParam )
{
	Critical
	global vMsgNum
	;; When the function is triggered (by a new window message), do nothing, thus also a new WE window is kept.
	;; Then (after doing nothing) set OnMessage back to the normal function.
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		OnMessage( vMsgNum, "ShellMessage_forCtrlWinE", 0 )
		OnMessage( vMsgNum, "ShellMessage" )
	}
}

ShellMessage( wParam, lParam )
{
	global
	Critical

/*wParam =
HSHELL_ACCESSIBILITYSTATE
11
The accessibility state has changed.

HSHELL_ACTIVATESHELLWINDOW
3
The shell should activate its main window.

HSHELL_APPCOMMAND
12
The user completed an input event (for example, pressed an application command button on the mouse or an application command key on the keyboard), and the application did not handle the WM_APPCOMMAND message generated by that input.
If the Shell procedure handles the WM_COMMAND message, it should not call CallNextHookEx. See the Return Value section for more information.

HSHELL_GETMINRECT
5
A window is being minimized or maximized. The system needs the coordinates of the minimized rectangle for the window.

HSHELL_LANGUAGE
8
Keyboard language was changed or a new keyboard layout was loaded.

HSHELL_REDRAW
6
The title of a window in the task bar has been redrawn.

HSHELL_TASKMAN
7
The user has selected the task list. A shell application that provides a task list should return TRUE to prevent Windows from starting its task list.

HSHELL_WINDOWACTIVATED
4
The activation has changed to a different top-level, unowned window.

HSHELL_WINDOWCREATED
1
A top-level, unowned window has been created. The window exists when the system calls this hook.

HSHELL_WINDOWDESTROYED
2
A top-level, unowned window is about to be destroyed. The window still exists when the system calls this hook.

HSHELL_WINDOWREPLACED
13
A top-level window is being replaced. The window exists when the system calls this hook. 

lParam differs in type according to the value of wParam received. For most of the wParam values, the lParam is a handle to a window that can be used as ahk_id %lParam% in AHK's Window commands.
*/

	; static vHWndsToIgnore := "|"

	; if ( (wParam = 1 or wParam = 6) and WinActive("ahk_class CabinetWClass"))
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		; if (InStr(vHWndsToIgnore, "|" lParam "|"))
		; 	return
		if (vKeepIfCtrlHeldDown and GetKeyState("Control", "P"))
		{
			; vHWndsToIgnore .= lParam "|"
			return
		}

		vPath := Explorer_GetPath(lParam)
		while ( vPath = "ERROR" and A_Index < 10)
		{
			 Sleep 100
			 vPath := Explorer_GetPath(lParam)
		}
		if ( vPath = "ERROR" )
		{
			 MsgBox Could not retrieve the path of the explorer window.
			 return
		}


		;; If RegEx is *not* used as compare method, then keep WE in case of the content of a zip file
		if (not vUseRegExForPathCompare)
		{
			if (StrLen(vPath) > 3 and StrLen(vPath) - InStr(vPath, ".zip", 0, 0) = 3)
				return
		}

		
		if (vUseRegExForPathCompare)
		{
			; MsgBox, % vPath ", " vWindowsExplorerPaths ", " vKeepOrRedirect_matchingWEPaths
			if (RegExMatch(vPath, vWindowsExplorerPaths))
			{
				if (vKeepOrRedirect_matchingWEPaths = "Keep" and (not vExceptionPathsInCaseOfRegEx or not RegExMatch(vPath, vExceptionPathsInCaseOfRegEx)))
				{
					; MsgBox, 1
					return
				}
				else if (vKeepOrRedirect_matchingWEPaths = "Redirect" and vExceptionPathsInCaseOfRegEx and RegExMatch(vPath, vExceptionPathsInCaseOfRegEx))
				{
					; MsgBox, 2
					return
				}
				; MsgBox, 3
			}
			else if (vKeepOrRedirect_matchingWEPaths = "Redirect")
					return
		}
		else if (InStr(vWindowsExplorerPaths, "|" . vPath . "|"))
		{
			if (vKeepOrRedirect_matchingWEPaths = "Keep")
				return
		}
		else if (vKeepOrRedirect_matchingWEPaths = "Redirect")
			return

			

		SetTitleMatchMode, 2
		IfWinNotExist, XYplorer ahk_class ThunderRT6FormDC
		{
			SplitPath, vPathToXYplorerForAutostart, OutFileName, OutDir, OutExtension, OutNameNoExt, OutDrive
			; Run, XYPlorer.exe, OutDir, UseErrorLevel
			Run, %vPathToXYplorerForAutostart%, OutDir, UseErrorLevel
			if (ErrorLevel = "ERROR")
			{
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					MsgBox, 262145, %A_ScriptName%, XYplorer could not be started automatically. Please, start it manually.`n`nOnce XYplorer is started you can close this dialog in order to redirect the Windows Explorer window to XYplorer.`n`nPress Cancel in order to *not* redirect the Windows Explorer window to XYplorer this time.
					IfMsgBox, Cancel
						return
				}
			}
			else
			{
				; WinWait, XYplorer ahk_class ThunderRT6FormDC
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					ToolTip, %A_ScriptName%:`n`nXYplorer is being started...
					Sleep, 20
				}
				ToolTip
			}
			IfWinNotExist, ahk_id %lParam%
				return			
		}

		WinWait, XYplorer ahk_class ThunderRT6FormDC
		WinActivate, XYplorer ahk_class ThunderRT6FormDC

		vSelected := Explorer_GetSelected(lParam)
		StringReplace, vSelected, vSelected, % Chr(10), |, 1

		WinClose, ahk_id %lParam%


		;; if vPath is found in a normal way in the file system and is *not* a Directory = vPath is a file (e.g. a zip file)
		vFileAttrib := FileExist(vPath)
		if (vFileAttrib and not InStr(vFileAttrib, "D"))
		{
			SplitPath, vPath, vOutFileName, vOutDir
			vSelected := vPath
			vPath := vOutDir
		}
		XY_send_message_noreturnvalue("::if(strpos('|' . get('tabs') . '|', ""|" . vPath . "|"") == -1) {tab('new', """ . vPath . """);} else {goto """ . vPath . """, 1;}")
		if (vSelected)
		{
			vArrSelected := StrSplit(vSelected, "|")
			XY_send_message_noreturnvalue("::selfilter row(""" . vArrSelected[1] . """) . ',' . get('CountItems'), , '#', 0, 1; selectitems """ . vSelected . """;")
		}
	}
}
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~   END of ShellMessage( wParam, lParam )   ~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



_TriggerWEwindowRedirectionToXY()
{
	global
	Critical
	if (not vHWndWE := WinActive("ahk_class CabinetWClass"))
		return
	vWindowsExplorerPaths_org := vWindowsExplorerPaths
	vWindowsExplorerPaths := ""
	vKeepOrRedirect_matchingWEPaths_org := vKeepOrRedirect_matchingWEPaths
	vKeepOrRedirect_matchingWEPaths := "Redirect"
	vUseRegExForPathCompare_org := vUseRegExForPathCompare
	vUseRegExForPathCompare := 1
	vExceptionPathsInCaseOfRegEx_org := vExceptionPathsInCaseOfRegEx
	vExceptionPathsInCaseOfRegEx := ""
	vKeepIfCtrlHeldDown_org := vKeepIfCtrlHeldDown
	vKeepIfCtrlHeldDown := 0
	ShellMessage( 1, vHWndWE )
	vWindowsExplorerPaths := vWindowsExplorerPaths_org
	vKeepOrRedirect_matchingWEPaths := vKeepOrRedirect_matchingWEPaths_org
	vUseRegExForPathCompare := vUseRegExForPathCompare_org
	vExceptionPathsInCaseOfRegEx := vExceptionPathsInCaseOfRegEx_org
	vKeepIfCtrlHeldDown := vKeepIfCtrlHeldDown_org
}


_ToggleUseControlToKeepWE()
{
	global
	vKeepIfCtrlHeldDown := vKeepIfCtrlHeldDown ? 0 : 1
	Hotkey, ^#e, Toggle
	Menu, Tray, ToggleCheck, Use [Ctrl] to manually keep a WE window
}


_ToggleHotkeyIWETTRTXY()
{
	global
	Menu, Tray, ToggleCheck, Use [%vDisplayFormatOfHKIWETTRTXY%] to manually redirect WE to XY
	Hotkey, IfWinActive, ahk_class CabinetWClass
	Hotkey, %vHotkeyInWEwindowsToTriggerRedirectionToXY%, Toggle
	Hotkey, If
}


_TogglePauseAndSuspendScript()
{
	global
	Suspend, Toggle
	if A_IsPaused
		OnMessage( vMsgNum, "ShellMessage")
	else
		OnMessage( vMsgNum, "ShellMessage", 0 )
	; MsgBox test2
	Pause, Toggle, 1 ;; The Pause only is used for the tray icon to show a red S instead of a green S. Otherwise, it is not necessary to use the Pause command. And for checking A_IsPaused in the if clause above. Putting "Suspend, Toggle" after the if clause, does not prevent the hotkey from being suspended. But if I wanted to check in the if clause for A_IsSuspended, then I would have to reverse the if-logic, checking for if *not* A_IsSuspended, since it is already after the Suspend command in the code, which is stupid from the logical part of what this function should do. Thus, since Pause also helps with the tray icon, I might as well use it in the if clause.
	Menu, Tray, ToggleCheck, Suspend Script ([%vDisplayFormatOfHKGTSUSTWS%])
}


_ExitApp()
{
	ExitApp
}


_WinGetClass(WinTitle := "", WinText := "", ExcludeTitle := "", ExcludeText := "")
{
	WinGetClass, class, %WinTitle%, %WinText%, %ExcludeTitle%, %ExcludeText%
	return class
}



;*****************************************************************************************************************
;~ Code from others  (or based on):
;*****************************************************************************************************************
/*
	Library for getting info from a specific explorer window (if window handle not specified, the currently active
	window will be used).  Requires AHK_L or similar.  Works with the desktop.  Does not currently work with save
	dialogs and such.
	
	
	Explorer_GetSelected(hwnd="")   - paths of target window's selected items
	Explorer_GetAll(hwnd="")        - paths of all items in the target window's folder
	Explorer_GetPath(hwnd="")       - path of target window's folder
	
	example:
		F1::
			path := Explorer_GetPath()
			all := Explorer_GetAll()
			sel := Explorer_GetSelected()
			MsgBox % path
			MsgBox % all
			MsgBox % sel
		return
	
	Joshua A. Kinnison
	2011-04-27, 16:12
*/

Explorer_GetPath(hwnd="")
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
		return A_Desktop
	path := window.LocationURL
	if (path="")
		path:=window.LocationName
	path := RegExReplace(path, "ftp://.*@","ftp://")
	StringReplace, path, path, file:///
	StringReplace, path, path, /, \, All 
	
	; thanks to polyethene
	Loop
		if RegExMatch(path, "i)(?<=%)[\da-f]{1,2}", hex)
			StringReplace, path, path, `%%hex%, % Chr("0x" . hex), All
		Else Break
	return path
}

Explorer_GetAll(hwnd="")
{
	return Explorer_Get(hwnd)
}

Explorer_GetSelected(hwnd="")
{
	return Explorer_Get(hwnd,true)
}

Explorer_GetWindow(hwnd="")
{
	; thanks to jethrow for some pointers here
    WinGet, process, processName, % "ahk_id " hwnd := hwnd? hwnd:WinExist("A")
    WinGetClass class, ahk_id %hwnd%


	; MsgBox, %hwnd%, %process%, %class%
	; ; if not process
	; 	MsgBox, % WinExist("ahk_id " hwnd)
	; ; return
	
	if (process!="explorer.exe")
		return
	if (class ~= "(Cabinet|Explore)WClass")
	{
		for window in ComObjCreate("Shell.Application").Windows
			try if (window.hwnd==hwnd)
				return window
	}
	else if (class ~= "Progman|WorkerW") 
		return "desktop" ; desktop found
}

Explorer_Get(hwnd="",selection=false)
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
	{
		ControlGet, hwWindow, HWND,, SysListView321, ahk_class Progman
		if !hwWindow ; #D mode
			ControlGet, hwWindow, HWND,, SysListView321, A
		ControlGet, files, List, % ( selection ? "Selected":"") "Col1",,ahk_id %hwWindow%
		base := SubStr(A_Desktop,0,1)=="\" ? SubStr(A_Desktop,1,-1) : A_Desktop
		Loop, Parse, files, `n, `r
		{
			path := base "\" A_LoopField
			IfExist %path% ; ignore special icons like Computer (at least for now)
				ret .= path "`n"
		}
	}
	else
	{
		if selection
			collection := window.document.SelectedItems
		else
			collection := window.document.Folder.Items
		for item in collection
			ret .= item.path "`n"
	}
	return Trim(ret,"`n")
}

;*****************************************************************************************************************

;heavily based on
;http://www.xyplorer.com/xyfc/viewtopic.php?f=7&t=9233&p=89555&hilit=messenger#p89555

XY_send_message_noreturnvalue(MessagetoXYplorer)          ;Send message to XYplorer
{
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  if !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
}

global XY_eval_and_returnvalue_XYRETURN

XY_eval_and_returnvalue(expression)          ;Send message to XYplorer
{
  global XY_eval_and_returnvalue_XYRETURN
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  MessagetoXYplorer := "::CopyData " SenderHWND ", " expression ", 0"    ;resolved to sth like this:   ::CopyData 7409230, <curitem>, 0   _OR like this:  ::CopyData 7409230, tab('get','count'), 0
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  if !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  OnMessage(0x4a, "Function_Receive_WM_COPYDATA")  ; 0x4a is WM_COPYDATA. This onhold and wait for the WM_Copydata from XYplorer then execute Function_Receive_WM_COPYDATA(wParam, lParam) below
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
  OnMessage(0x4a, "")
  return XY_eval_and_returnvalue_XYRETURN
}

Function_Receive_WM_COPYDATA(wParam, lParam)
{
  global XY_eval_and_returnvalue_XYRETURN
  StringAddress := NumGet(lParam + 2*A_PtrSize) ;lParam+8 is the address of CopyDataStruct's lpData member.
  CopyOfData := StrGet(StringAddress)    ;May also specify CP0 (default) or UTF-8 or UTF-16:   StrGet(StringAddress, NumGet(lParam+A_PtrSize), "UTF-16")
  cbData := NumGet(lParam+A_PtrSize)/2  ;cbData/2 = String length
  StringLeft, Datareceived, CopyOfData, cbData
  XY_eval_and_returnvalue_XYRETURN := Datareceived
}

;*****************************************************************************************************************
